// ==============================================================
//                 ORBITER MODULE: marsobserver
//                  
//
// marsobserver.cpp
// Control module for marsobserver vessel class
//
// 
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
#include <cstdio>
#include <cmath>
#include <string>
#include <stdio.h>
#include <math.h>

// ==============================================================
// Some vessel parameters
// ==============================================================

const double ISP = 3106;
const double RCSTH = 22;
const double MAINTH = 1960;
const double FUEL_MASS = 1346;
const double EMP_MASS = 1028;
const double BOOM_SPEED = 0.05;
const double ARRAY_SPEED = 0.025;
const double ARRAY_ROT_SPEED = 0.025;
const double ANT_DEPLOY_SPEED = 0.025;
const double ANT_ROT_SPEED = 0.025;


#define YELLOW RGB(255, 255, 80)
#define GREEN RGB(0, 225, 0)

const VECTOR3 lid_off = _V(0.684,0.482,1.24);



// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
void VLiftCoeff (VESSEL *v, double aoa, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD, -2*RAD, 10*RAD,15*RAD,50*RAD,90*RAD,180*RAD};
	static const double CL[nabsc]  = {0,0,0,0,0,0,0,0,0};
	static const double CM[nabsc]  = {0,0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

void HLiftCoeff (VESSEL *v, double beta, double M, double Re, void *context, double *cl, double *cm, double *cd)
{
	int i;
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {0,0,0,0,0,0,0,0};
	for (i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.95 + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}


class marsobserver: public VESSEL3 {
public:
	
	marsobserver (OBJHANDLE hVessel, int flightmodel);
	~marsobserver ();

	void clbkSetClassCaps (FILEHANDLE cfg);
	void DefineAnimations (void);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	//void clbkPostStep (double simt, double simdt, double mjd);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void JettisonLid(void);
	void DoKillRot(void);
	void GoBang(void);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey(char *kstate);
	int MyID;
	THRUSTER_HANDLE th_rcs[4], th_group[4], th_main[1];
	PROPELLANT_HANDLE hProp;
	ANIMATIONCOMPONENT_HANDLE Section0, Section1, Section2, Section3, Section4, Section5, Section6, Section7;
	AIRFOILHANDLE hHA, hVA;

private:
	UINT anim_ant_deploy,anim_ant_rota, anim_ant_rotb, anim_solp_deploy, anim_solp_rota, anim_solp_rotb;
	UINT anim_grs_boom, anim_mag_boom;
	double ant_deploy_proc, ant_rota_proc, ant_rotb_proc, solp_deploy_proc, solp_rota_proc, solp_rotb_proc;
	double grs_boom_proc, mag_boom_proc, timer, pstream_lvl;
	int  ant_status, solp_status, lid_status, kilrot_sw, bang_sw, boom_status;


};

marsobserver::marsobserver (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{

}

marsobserver::~marsobserver ()
{
	
}

// Define animation sequences for moving parts
// --------------------------------------------------------------
void marsobserver::DefineAnimations (void)
{


	
	// 1. Solar arrays
	anim_solp_deploy = CreateAnimation (0);
	anim_solp_rota = CreateAnimation (0);
	anim_solp_rotb = CreateAnimation (0);

	static UINT ArrayCFoldGrp1[1] = {48};
	static UINT ArrayCFoldGrp2[1] = {12};

	static UINT ArrayLFoldGrp1[1] = {51};
	static UINT ArrayLFoldGrp2[1] = {52};

	static UINT ArrayRFoldGrp1[1] = {50};
	static UINT ArrayRFoldGrp2[1] = {49};

	static UINT ArrayPivGrp1[2] = {19,39};

	static UINT ArrayArmGrp1[2] = {20,23};
	static UINT ArrayArmGrp2[2] = {21,22};
	



	static MGROUP_ROTATE Arm1 (0, ArrayArmGrp1, 2, _V(0,0.679,0.49), _V(1,0,0), (float)(90*RAD));
	Section0 = AddAnimationComponent (anim_solp_deploy, 0.65, 1.0, &Arm1);

	static MGROUP_ROTATE Arm2 (0, ArrayArmGrp2, 2, _V(0,0.703,-0.555), _V(1,0,0), (float)(-180*RAD));
	Section1 = AddAnimationComponent (anim_solp_deploy, 0.65, 1.0, &Arm2, Section0);

	static MGROUP_ROTATE Piv1 (0, ArrayPivGrp1, 2, _V(0,0.727,0), _V(0,0,1), (float)(360*RAD));
	Section2 = AddAnimationComponent (anim_solp_rota, 0.0,  1, &Piv1, Section1);

	static MGROUP_ROTATE Piv2 (0, ArrayCFoldGrp1, 1, _V(0,0.727,0.807), _V(1,0,0), (float)(90*RAD));
	Section3 = AddAnimationComponent (anim_solp_rotb, 0.0,  1, &Piv2, Section2);

	static MGROUP_ROTATE Sp1 (0, ArrayCFoldGrp2, 1, _V(0,0.794,0.806), _V(1,0,0), (float)(180*RAD));
	AddAnimationComponent (anim_solp_deploy, 0.5,  0.7, &Sp1, Section3);

	static MGROUP_ROTATE Sp2 (0, ArrayLFoldGrp1, 1, _V(1.142,0.821,0), _V(0,0,1), (float)(-180*RAD));
	Section4 = AddAnimationComponent (anim_solp_deploy, 0.0,  0.2, &Sp2, Section3);

	static MGROUP_ROTATE Sp3 (0, ArrayLFoldGrp2, 1, _V(0,0.848,0.816), _V(1,0,0), (float)(-180*RAD));
	AddAnimationComponent (anim_solp_deploy, 0.2,  0.4, &Sp3, Section4);

	static MGROUP_ROTATE Sp4 (0, ArrayRFoldGrp1, 1, _V(-1.142,0.807,0), _V(0,0,1), (float)(180*RAD));
	Section5 = AddAnimationComponent (anim_solp_deploy, 0.1,  0.3, &Sp4, Section3);

	static MGROUP_ROTATE Sp5 (0, ArrayRFoldGrp2, 1, _V(0,0.821,0.816), _V(1,0,0), (float)(-180*RAD));
	AddAnimationComponent (anim_solp_deploy, 0.3,  0.5, &Sp5, Section5);

	

	// 3. ant deploy
	anim_ant_deploy = CreateAnimation (0);

	static UINT AntGrp11[3] = {17,18,32};
	static MGROUP_ROTATE Ant11 (0, AntGrp11, 3, _V(0,-0.68,-0.499), _V(1,0,0), (float)(180*RAD));
	Section0 = AddAnimationComponent (anim_ant_deploy, 0.5, 1.0, &Ant11);

	static UINT AntGrp1[3] = {15,16,31};
	static MGROUP_ROTATE Ant1 (0, AntGrp1, 3, _V(0,-0.71,2.079), _V(1,0,0), (float)(-180*RAD));
	Section1 = AddAnimationComponent (anim_ant_deploy, 0.0, 0.5, &Ant1, Section0);

	

	// 4. ant rot
	anim_ant_rota = CreateAnimation (0);

	static UINT AntGrp2[1] = {13};
	static MGROUP_ROTATE Ant2 (0, AntGrp2, 1, _V(0,-0.734,0), _V(0,0,1), (float)(360*RAD));
	Section2 = AddAnimationComponent (anim_ant_rota, 0.0, 1, &Ant2, Section1);

	anim_ant_rotb = CreateAnimation (0);

	static UINT AntGrp23[4] = {14,28,29,30};
	static MGROUP_ROTATE Ant23 (0, AntGrp23, 0, _V(0,-0.745,-0.549), _V(1,0,0), (float)(180*RAD));
	Section3 = AddAnimationComponent (anim_ant_deploy, 0.3, 0.5, &Ant23, Section2);

	static UINT AntGrp22[4] = {14,28,29,30};
	static MGROUP_ROTATE Ant22 (0, AntGrp22, 4, _V(0,-0.745,-0.549), _V(1,0,0), (float)(-90*RAD));
	AddAnimationComponent (anim_ant_rotb, 0.0, 1, &Ant22, Section3);

	////// booms - grs
	anim_grs_boom = CreateAnimation (0);

	static UINT grsgrp0[2] = {7,42};
	static MGROUP_ROTATE grs0 (0, grsgrp0, 2, _V(-1.281,0,-0.457), _V(0,1,0), (float)(-90*RAD));
	Section0 = AddAnimationComponent (anim_grs_boom, 0.2, 0.4, &grs0);

	static UINT grsgrp2[1] = {6};
	static MGROUP_TRANSLATE grs2 (0, grsgrp2, 1,  _V(0,0,1.2));
	Section1 = AddAnimationComponent (anim_grs_boom, 0.8, 1.0, &grs2, Section0);

	static UINT grsgrp3[1] = {40};
	static MGROUP_TRANSLATE grs3 (0, grsgrp3, 1,  _V(0,0,1.2));
	Section2 = AddAnimationComponent (anim_grs_boom, 0.6, 0.8, &grs3, Section1);

	static UINT grsgrp4[1] = {41};
	static MGROUP_TRANSLATE grs4 (0, grsgrp4, 1,  _V(0,0,1.2));
	Section3 = AddAnimationComponent (anim_grs_boom, 0.4, 0.6, &grs4, Section2);

	static UINT grsgrp1[4] = {5,8,9,11};
	static MGROUP_TRANSLATE grs1 (0, grsgrp1, 4,  _V(0,0,1.2));
	Section4 = AddAnimationComponent (anim_grs_boom, 0.0, 0.2, &grs1, Section3);

	static UINT grsgrp5[1] = {10};
	static MGROUP_ROTATE grs5 (0, grsgrp5, 1, _V(-1.137,-0.246,0), _V(0,0,1), (float)(-180*RAD));
	AddAnimationComponent (anim_grs_boom, 0.1, 0.2, &grs5, Section4);

	// mag boom
	anim_mag_boom = CreateAnimation (0);
	
	static UINT maggrp2[1] = {2};
	static MGROUP_TRANSLATE mag2 (0, maggrp2, 1,  _V(1.19776,-0.07326,0));
	Section0 = AddAnimationComponent (anim_mag_boom, 0.8, 1.0, &mag2);

	static UINT maggrp3[1] = {43};
	static MGROUP_TRANSLATE mag3 (0, maggrp3, 1,  _V(1.19776,-0.07326,0));
	Section1 = AddAnimationComponent (anim_mag_boom, 0.6, 0.8, &mag3, Section0);

	static UINT maggrp4[1] = {44};
	static MGROUP_TRANSLATE mag4 (0, maggrp4, 1,  _V(1.19776,-0.07326,0));
	Section2 = AddAnimationComponent (anim_mag_boom, 0.4, 0.6, &mag4, Section1);

	static UINT maggrp5[1] = {45};
	static MGROUP_TRANSLATE mag5 (0, maggrp5, 1,  _V(1.19776,-0.07326,0));
	Section3 = AddAnimationComponent (anim_mag_boom, 0.2, 0.4, &mag5, Section2);

	static UINT maggrp6[3] = {0,1,46};
	static MGROUP_TRANSLATE mag6 (0, maggrp6, 3,  _V(1.19776,-0.07326,0));
	AddAnimationComponent (anim_mag_boom, 0.0, 0.2, &mag6, Section3);

	/// pmir door link to grs door
	static UINT grsgrp6[1] = {26};
	static MGROUP_ROTATE grs6 (0, grsgrp6, 1, _V(0,-0.764,0.873), _V(1,0,0), (float)(-90*RAD));
	AddAnimationComponent (anim_grs_boom, 0.1, 0.2, &grs6);


}



// ==============================================================
// marsobserver class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void marsobserver::clbkSetClassCaps (FILEHANDLE cfg)
{
	ant_deploy_proc = 0.0;
	ant_rota_proc = 0.0;
	ant_rotb_proc = 0.0;
	solp_deploy_proc = 0.0;
	solp_rota_proc = 0.0;
	solp_rotb_proc = 0.0;
	grs_boom_proc = 0.0;
	mag_boom_proc = 0.0;
	ant_status = 0;
	solp_status = 0;
	lid_status = 0;
	boom_status = 0;
	kilrot_sw = 0;
	bang_sw = 0;
	timer = 0.0;
	pstream_lvl = 0.0;
	
	DefineAnimations ();
	// physical specs
	SetCW (1, 1, 1, 1);
	SetRotDrag (_V(1,1,1));
	SetPMI (_V(1,2,2));
	SetCrossSections (_V(1,2,2));
	SetSize (2);
	SetEmptyMass (EMP_MASS);
	SetMeshVisibilityMode ((AddMesh ("MarsObserver/marsobserver")), MESHVIS_EXTERNAL);
	SetMeshVisibilityMode ((AddMesh ("MarsObserver/moc_lid", &lid_off)), MESHVIS_EXTERNAL);
	
	SetCameraOffset (_V(0,0,0));
	SetTouchdownPoints (_V(0,0.8,-1.8), _V(0.512,-0.853,-1.8), _V(-0.512,-0.853,-1.8));
	SetSurfaceFrictionCoeff (1,1);
	SetDockParams (_V(0,0,-2), _V(0,0,-1), _V(0,1,0));

	// propellant resources

	hProp = CreatePropellantResource (FUEL_MASS, FUEL_MASS);

	
	// ***************** thruster definitions *******************
	th_main[0] = CreateThruster (_V( 0,0,0), _V(0,0, 1), MAINTH, hProp, ISP);
	
	SURFHANDLE mainex = oapiRegisterExhaustTexture ("Exhaust2");
	AddExhaust (th_main[0], 3, 0.2, _V( 0.814,0.416,-1.04), _V(0,0,-1), mainex);
	AddExhaust (th_main[0], 3, 0.2, _V( -0.814,0.416,-1.04), _V(0,0,-1), mainex);
	AddExhaust (th_main[0], 3, 0.2, _V( 0.814,-0.416,-1.04), _V(0,0,-1), mainex);
	AddExhaust (th_main[0], 3, 0.2, _V( -0.814,-0.416,-1.04), _V(0,0,-1), mainex);
	
	
	
	
	
	
	th_rcs[ 0] = CreateThruster (_V(1.143, 0.671, -1.01), _V(0,0,1), RCSTH, hProp, ISP);
	th_rcs[ 1] = CreateThruster (_V(-1.143, 0.671, -1.01), _V(0,0,1), RCSTH, hProp, ISP);
	th_rcs[ 2] = CreateThruster (_V(1.143, -0.671, -1.01), _V(0,0,1), RCSTH, hProp, ISP);
	th_rcs[ 3] = CreateThruster (_V(-1.143,-0.671, -1.01), _V(0,0,1), RCSTH, hProp, ISP);
	

	SURFHANDLE tex_rcs = oapiRegisterExhaustTexture ("exhaust_atrcs");
	for(int i=0;i<4;i++)
	{
	AddExhaust (th_rcs[i], 2, 0.1, (tex_rcs));
	}

	th_group[0] = th_rcs[ 2];
	th_group[1] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 1];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 2];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[ 1];
	th_group[1] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[ 0];
	th_group[1] = th_rcs[ 1];
	th_group[2] = th_rcs[ 2];
	th_group[3] = th_rcs[ 3];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_FORWARD);

	th_group[0] = th_main[ 0];
	CreateThrusterGroup (th_group, 1, THGROUP_MAIN);



	///////////////////////////////////////////////////////
	/////AirFoils


	hVA = CreateAirfoil3 (LIFT_VERTICAL, _V(0,0,0.1), VLiftCoeff, 0, 2, 2, 1);
	hHA = CreateAirfoil3 (LIFT_HORIZONTAL, _V(0,0,0.1), HLiftCoeff, 0, 2, 2, 1);

	////////////////////// boom pstreams

	SURFHANDLE exh = oapiRegisterExhaustTexture ("contrail3");

	PARTICLESTREAMSPEC big_pstream = {
		0, 2, 5, 10, 0.7, 5, 20, 4.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 8,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1
		};

	PARTICLESTREAMSPEC small_pstream = {
		0, 0.25, 1, 25, 2, 2, 0, 4.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
		exh};

	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,1,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(0,-1,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(1,0,0), &pstream_lvl);
	AddParticleStream(&small_pstream, _V(0,0,0), _V(-1,0,0), &pstream_lvl);

	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,0,1), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,0,-1), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,1,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(0,-1,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(1,0,0), &pstream_lvl);
	AddParticleStream(&big_pstream, _V(0,0,0), _V(-1,0,0), &pstream_lvl);

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new marsobserver (hvessel, flightmodel);
}




// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (marsobserver*)vessel;
}
// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void marsobserver::clbkPostCreation ()
{	
	MyID=ConnectToOrbiterSoundDLL(GetHandle());
	ReplaceStockSound(MyID,"Sound\\Vessel\\attsustain.wav",	REPLACE_MAIN_THRUST);

	if(lid_status == 1) DelMesh(1);

	if (GetPropellantMass(hProp)< 0.01 && oapiGetSimTime ()> 0.1)SetPropellantMass(hProp, FUEL_MASS);

	if(GetDamageModel() == 1) 
	{
		bang_sw = 1;
		timer = 10 + (40*oapiRand());
	}
}
void marsobserver::clbkSaveState (FILEHANDLE scn)
{

	char cbuf[256];

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	// custom parameters

	sprintf (cbuf, "%d %0.4f %0.4f %0.4f", ant_status, ant_deploy_proc, ant_rota_proc, ant_rotb_proc);
	oapiWriteScenario_string (scn, "ANT", cbuf);

	sprintf (cbuf, "%d %0.4f %0.4f %0.4f %d %0.4f %0.4f", solp_status, solp_deploy_proc, solp_rota_proc, solp_rotb_proc,boom_status, grs_boom_proc, mag_boom_proc);
	oapiWriteScenario_string (scn, "SLP", cbuf);

	sprintf (cbuf, "%d", lid_status);
	oapiWriteScenario_string (scn, "SCH", cbuf);
	
}
void marsobserver::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "ANT", 3)) {
            sscanf (line+3, "%d%lf%lf%lf", &ant_status, &ant_deploy_proc, &ant_rota_proc, &ant_rotb_proc);
		}else if (!strnicmp (line, "SLP", 3)) {
            sscanf (line+3, "%d%lf%lf%lf%d%lf%lf", &solp_status, &solp_deploy_proc, &solp_rota_proc, &solp_rotb_proc, &boom_status, &grs_boom_proc, &mag_boom_proc);
		}else if (!strnicmp (line, "SCH", 3)) {
			sscanf_s (line+3, "%d", &lid_status);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
   }
	SetAnimation (anim_ant_deploy, ant_deploy_proc);
	SetAnimation (anim_ant_rota, ant_rota_proc);
	SetAnimation (anim_ant_rotb, ant_rotb_proc);
	
	SetAnimation (anim_solp_deploy, solp_deploy_proc);
	SetAnimation (anim_solp_rota, solp_rota_proc);
	SetAnimation (anim_solp_rotb, solp_rotb_proc);

	SetAnimation (anim_grs_boom, grs_boom_proc);
	SetAnimation (anim_mag_boom, mag_boom_proc);
}






int marsobserver::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {

  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) 
	{
	  switch (key) 
	  {
		case OAPI_KEY_NUMPAD5: 
		if  (kilrot_sw == 0) kilrot_sw = 1;
		else kilrot_sw = 0;
		return 1;
	  }
	  return true;
	} 
  else if (KEYMOD_CONTROL (kstate)) 
	{
  
	}
  else 
	{ // unmodified keys
    switch (key) 
		{
	case OAPI_KEY_G: 
		if (solp_status  == 0)solp_status = 1;
		else if (solp_status  == 2)solp_status = 3;
		if (ant_status == 0)ant_status = 1;
		else if (ant_status == 2)ant_status = 3;
		if (lid_status == 0)JettisonLid();
		return 1;

		case OAPI_KEY_N: 
		if (boom_status == 0)boom_status = 1;
		else if (boom_status == 1)boom_status = 3;
		else if (boom_status == 2)boom_status = 3;
		else if (boom_status == 3)boom_status = 1;
		else if (boom_status == 4)boom_status = 1;
		else if (boom_status == 5)boom_status = 1;
		return 1;

		case OAPI_KEY_M: 
		if (boom_status == 0)boom_status = 4;
		else if (boom_status == 1)boom_status = 4;
		else if (boom_status == 2)boom_status = 4;
		else if (boom_status == 3)boom_status = 4;
		else if (boom_status == 4)boom_status = 3;
		else if (boom_status == 5)boom_status = 3;
		return 1;

		/*case OAPI_KEY_J: 
		if (lid_status == 0)JettisonLid();
		return 1;*/
		}
	}
  
 return 0;
 }


int marsobserver::clbkConsumeDirectKey(char *kstate)
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

if (KEYMOD_SHIFT (kstate))
{
if (KEYDOWN (kstate, OAPI_KEY_NUMPAD2))
		{
		if (avel.x < RAD*10)
			{
			avel.x = avel.x+drot;
			if (avel.x > RAD*10) avel.x = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD8))
		{
		if (avel.x > -RAD*10)
			{
			avel.x = avel.x-drot;
			if (avel.x < -RAD*10) avel.x = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD1))
		{
		if (avel.y < RAD*10)
			{
			avel.y = avel.y+drot;
			if (avel.y > RAD*10) avel.y = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD3))
		{
		if (avel.y > -RAD*10)
			{
			avel.y = avel.y-drot;
			if (avel.y < -RAD*10) avel.y = -RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD6))
		{
		if (avel.z < RAD*10)
			{			
			avel.z = avel.z+drot;
			if (avel.z > RAD*10) avel.z = RAD*10;
			}
		}

if (KEYDOWN (kstate, OAPI_KEY_NUMPAD4))
		{
		if (avel.z > -RAD*10)
			{
			avel.z = avel.z-drot;
			if (avel.z < -RAD*10) avel.z = -RAD*10;
			}		
		}

		SetAngularVel(avel);
	}
return 0;
}


void marsobserver::DoKillRot()
{
		VECTOR3 avel = _V(0,0,0);
		GetAngularVel(avel);
		double dt = oapiGetSimStep();
		double drot = RAD*2*dt;

		if (drot > abs(avel.x)) avel.x = 0;
		else avel.x = avel.x + (drot * ((abs(avel.x))/-avel.x));
		if (drot > abs(avel.y)) avel.y = 0;
		else avel.y = avel.y + (drot * ((abs(avel.y))/-avel.y));
		if (drot > abs(avel.z)) avel.z = 0;
		else avel.z = avel.z + (drot * ((abs(avel.z))/-avel.z));

		SetAngularVel (avel);
		if (avel.x == 0 && avel.y == 0 && avel.z == 0) kilrot_sw = 0;
}


bool marsobserver::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int s = hps->H;
	double d = (s*0.00130208);
	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}



	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	//int w6 = (int)(280*d);
	//int w7 = (int)(296*d);
	//int w8 = (int)(312*d);
	//int w9 = (int)(328*d);
	//int w10 = (int)(344*d);
	//int w11 = (int)(360*d);
	//int w12 = (int)(376*d);
	//int w13 = (int)(392*d);
	//int w14 = (int)(408*d);
	//int w15 = (int)(424*d);

	{
	skp->SetTextColor(0x0066FF66);
	double m1 = GetEmptyMass();
	double m2 = GetPropellantMass(hProp);
	double mass_fraction = ((m1+m2)/(m1));
	double dv = (ISP*(log(mass_fraction)));
	char cbuf[256];
	sprintf(cbuf,"Current Available dV %0.2f m/s", dv);
	skp->Text ((10+roxl), (wc+royl), cbuf,strlen(cbuf));
	}


	{
	skp->SetTextColor(0x0066FF66);
	char hbuf[256];
	sprintf(hbuf,"Shift+NumPad = Reaction Wheel Att.Control");
	skp->Text ((10+roxl), (w1+royl), hbuf,strlen(hbuf));
	}

	if (kilrot_sw == 1)
	{
	skp->SetTextColor(0x0066FFFF);
	char gbuf[256];
	sprintf(gbuf,"                                          Kill Rot.Active");
	skp->Text ((10+roxl), (w1+royl), gbuf,strlen(gbuf));
	}

	if (solp_status == 0)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Sol.Panels/HGA Cruise Config.");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	else if (solp_status == 2)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"G = Sol.Panels/HGA Cruise Config.");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	if (boom_status > 0 && boom_status < 3)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"N = GRS/Mag.Boom Stow");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"N = GRS/Mag.Boom Cruise Config.");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}

	if (boom_status < 4)
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"M = GRS/Mag.Boom Mapping Config.");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}
	else
	{
	skp->SetTextColor(0x0066FF66);
	char abuf[256];
	sprintf(abuf,"M = GRS/Mag.Boom Stow.");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}


	if (bang_sw == 1)
	{
	skp->SetTextColor(0x006666FF);
	char abuf[256];
	sprintf(abuf,"Failure Mode ON");
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	}


	


	

	return true;
}


void marsobserver::clbkPreStep(double SimT, double SimDT, double MJD) {
	//sprintf(oapiDebugString(), "ps fm %d", GetFlightModel());

	if (kilrot_sw == 1) DoKillRot();
	if (GetThrusterGroupLevel(THGROUP_MAIN) > 0)
	{
		if (bang_sw == 1 && GetPropellantMass(hProp) < 1100) 
		{
		timer = timer - SimDT;
		if (timer < 0) 
			{
			timer = 0;
			bang_sw = 2;
			oapiSetTimeAcceleration(1);
			oapiCameraAttach (GetHandle(), 1);
			}
		}
	}
	if (bang_sw == 2) GoBang();
	///////// animations

	// deploy solp
	if (solp_status == 0) solp_deploy_proc = 0;
	else if (solp_status == 1) 
	{
		double da = SimDT * ARRAY_SPEED;
		solp_deploy_proc = solp_deploy_proc + da;
		if (solp_deploy_proc > 0.5) 
		{
			solp_deploy_proc = 0.5;
			solp_status = 2;
		}
	}
	else if (solp_status == 2)
	{
		solp_deploy_proc = 0.5;
	}
	else if (solp_status == 3)
	{
		double da = SimDT * ARRAY_SPEED;
		solp_deploy_proc = solp_deploy_proc + da;
		if (solp_deploy_proc > 1) 
		{
			solp_deploy_proc = 1;
			solp_status = 4;
		}
	}
	else if (solp_status == 4)
	{
		solp_deploy_proc = 1;
	
	//////// suntracking
	VECTOR3 sun_pos_glob = _V(0,0,0);
	VECTOR3 sun_pos_loc = _V(0,0,0);

	// rotb
	Global2Local(sun_pos_glob, sun_pos_loc);
	VECTOR3 sun_pos_locn = sun_pos_loc;
	normalise(sun_pos_locn);
	double sun_angle = acos(dotp(sun_pos_locn,_V(0,-1,0)));
	sun_angle = sun_angle-PI05;
	if(sun_angle < 0) sun_angle = 0;
	else if(sun_angle > PI05) sun_angle = PI05;

	double sun_angle_proc = sun_angle/PI05;

	if (solp_rotb_proc < sun_angle_proc)
	{
	double da = SimDT * ARRAY_ROT_SPEED;
	solp_rotb_proc = solp_rotb_proc+da;
	if (solp_rotb_proc > sun_angle_proc)solp_rotb_proc = sun_angle_proc;
	}

	else if (solp_rotb_proc > sun_angle_proc)
	{
	double da = SimDT * ARRAY_ROT_SPEED;
	solp_rotb_proc = solp_rotb_proc-da;
	if (solp_rotb_proc < sun_angle_proc)solp_rotb_proc = sun_angle_proc;
	}

	/////// rota
	sun_angle = atan2(-sun_pos_loc.x, -sun_pos_loc.z);
	if(sun_angle < 0) sun_angle = sun_angle + PI2;
	else if(sun_angle > PI2) sun_angle = sun_angle - PI2;

	sun_angle_proc = sun_angle/PI2;

	int rot_dir = 1;
	double proc_diff = (sun_angle_proc - solp_rota_proc);
	if (proc_diff > 0.5) rot_dir = -1;
	if (proc_diff < 0) rot_dir = -1;
	if (proc_diff < -0.5) rot_dir = 1;

	double da = SimDT * ARRAY_ROT_SPEED;
	if (proc_diff > 0.5)proc_diff = proc_diff - 1; // PI PI2
	else if (proc_diff < -0.5)proc_diff = proc_diff + 1;
	
	if (da > abs(proc_diff)) solp_rota_proc = sun_angle_proc;
	else solp_rota_proc = solp_rota_proc + (da*rot_dir);
	
	int srpi = (int) (solp_rota_proc);
	if (solp_rota_proc > 1) solp_rota_proc = solp_rota_proc - srpi;
	else if (solp_rota_proc < 0) solp_rota_proc = solp_rota_proc + 1 - srpi;
	
	}
	SetAnimation (anim_solp_deploy, solp_deploy_proc);
	SetAnimation (anim_solp_rota, solp_rota_proc);
	SetAnimation (anim_solp_rotb, solp_rotb_proc);



	/////////////////////////////////////////////////////// ant
	if (ant_status == 0) ant_deploy_proc = 0;
	else if (ant_status == 1) 
	{
		double da = SimDT * ANT_DEPLOY_SPEED;
		ant_deploy_proc = ant_deploy_proc + da;
		if (ant_deploy_proc > 0.5) 
		{
			ant_deploy_proc = 0.5;
			ant_status = 2;
		}
	}
	else if (ant_status == 2) 
	{
	ant_deploy_proc = 0.5;
	}
	else if (ant_status == 3) 
	{
	double da = SimDT * ANT_DEPLOY_SPEED;
		ant_deploy_proc = ant_deploy_proc + da;
		if (ant_deploy_proc > 1) 
		{
			ant_deploy_proc = 1;
			ant_status = 4;
		}
		
	}
	else if (ant_status == 4) 
	{
	ant_deploy_proc = 1;
	}
	SetAnimation (anim_ant_deploy, ant_deploy_proc);

	////////////////////////////////// earth tracking
	if (ant_status == 2) 
	{
	
	VECTOR3 earth_pos_glob = _V(0,0,0);
	VECTOR3 earth_pos_loc = _V(0,0,-1);
	OBJHANDLE hEarth = oapiGetObjectByName("Earth");
	if (hEarth)
	{
		oapiGetGlobalPos(hEarth, &earth_pos_glob);
		Global2Local(earth_pos_glob, earth_pos_loc);
	}
	normalise(earth_pos_loc);
	
	double earth_angle = atan2(earth_pos_loc.x, earth_pos_loc.y);
	earth_angle = earth_angle+PI;
	if (earth_angle < 0)earth_angle = earth_angle+PI2;
	else if (earth_angle > PI2)earth_angle = earth_angle-PI2;
	//sprintf(oapiDebugString(), "req_shaft_axis %0.4f %0.4f %0.4f", req_shaft_axis.x, req_shaft_axis.y, req_shaft_axis.z);

	double earth_angle_proc = earth_angle/PI2;

	int rot_dira = 1;
	double proc_diffa = (earth_angle_proc - ant_rota_proc);
	if (proc_diffa > 0.5) rot_dira = -1;
	if (proc_diffa < 0) rot_dira = -1;
	if (proc_diffa < -0.5) rot_dira = 1;

	double da = SimDT * ANT_ROT_SPEED;
	if (proc_diffa > 0.5)proc_diffa = proc_diffa - 1; // PI PI2
	else if (proc_diffa < -0.5)proc_diffa = proc_diffa + 1;
	
	if (da > abs(proc_diffa)) ant_rota_proc = earth_angle_proc;
	else ant_rota_proc = ant_rota_proc + (da*rot_dira);
	
	int srpi = (int) (ant_rota_proc);
	if (ant_rota_proc > 1) ant_rota_proc = ant_rota_proc - srpi;
	else if (ant_rota_proc < 0) ant_rota_proc = ant_rota_proc + 1 - srpi;
	//
	normalise(earth_pos_loc);
	earth_angle = acos(dotp(earth_pos_loc, _V(0,0,-1)));
	earth_angle = earth_angle-PI05;
	if (earth_angle < 0)earth_angle = 0;
	else if (earth_angle > PI05)earth_angle = PI05;
	
	earth_angle_proc = earth_angle/PI05;

	if (ant_rotb_proc < earth_angle_proc)
	{
		double da = SimDT*ANT_ROT_SPEED*4;
		ant_rotb_proc = ant_rotb_proc+da;
		if (ant_rotb_proc > earth_angle_proc)ant_rotb_proc = earth_angle_proc;
	}
	else if (ant_rotb_proc > earth_angle_proc)
	{
		double da = SimDT*ANT_ROT_SPEED*4;
		ant_rotb_proc = ant_rotb_proc-da;
		if (ant_rotb_proc < earth_angle_proc)ant_rotb_proc = earth_angle_proc;
	}

	SetAnimation (anim_ant_rota, ant_rota_proc);
	SetAnimation (anim_ant_rotb, ant_rotb_proc);
	}
	//////////////////////////////////////////
	else if (ant_status == 4) 
	{
	
	VECTOR3 earth_pos_glob = _V(0,0,0);
	VECTOR3 earth_pos_loc = _V(0,0,-1);
	OBJHANDLE hEarth = oapiGetObjectByName("Earth");
	if (hEarth)
	{
		oapiGetGlobalPos(hEarth, &earth_pos_glob);
		Global2Local(earth_pos_glob, earth_pos_loc);
	}
	normalise(earth_pos_loc);
	
	double earth_angle = atan2(earth_pos_loc.x, -earth_pos_loc.y);
	earth_angle = earth_angle+PI;
	
	if (earth_angle < 0)earth_angle = earth_angle+PI2;
	else if (earth_angle > PI2)earth_angle = earth_angle-PI2;
	//sprintf(oapiDebugString(), "req_shaft_axis %0.4f %0.4f %0.4f", req_shaft_axis.x, req_shaft_axis.y, req_shaft_axis.z);

	double earth_angle_proc = earth_angle/PI2;

	int rot_dira = 1;
	double proc_diffa = (earth_angle_proc - ant_rota_proc);
	if (proc_diffa > 0.5) rot_dira = -1;
	if (proc_diffa < 0) rot_dira = -1;
	if (proc_diffa < -0.5) rot_dira = 1;

	double da = SimDT * ANT_ROT_SPEED;
	if (proc_diffa > 0.5)proc_diffa = proc_diffa - 1; // PI PI2
	else if (proc_diffa < -0.5)proc_diffa = proc_diffa + 1;
	
	if (da > abs(proc_diffa)) ant_rota_proc = earth_angle_proc;
	else ant_rota_proc = ant_rota_proc + (da*rot_dira);
	
	int srpi = (int) (ant_rota_proc);
	if (ant_rota_proc > 1) ant_rota_proc = ant_rota_proc - srpi;
	else if (ant_rota_proc < 0) ant_rota_proc = ant_rota_proc + 1 - srpi;
	//
	normalise(earth_pos_loc);
	earth_angle = acos(dotp(earth_pos_loc, _V(0,0,1)));
	earth_angle = earth_angle-PI05;
	if (earth_angle < 0)earth_angle = 0;
	else if (earth_angle > PI05)earth_angle = PI05;
	
	earth_angle_proc = earth_angle/PI05;

	if (ant_rotb_proc < earth_angle_proc)
	{
		double da = SimDT*ANT_ROT_SPEED*4;
		ant_rotb_proc = ant_rotb_proc+da;
		if (ant_rotb_proc > earth_angle_proc)ant_rotb_proc = earth_angle_proc;
	}
	else if (ant_rotb_proc > earth_angle_proc)
	{
		double da = SimDT*ANT_ROT_SPEED*4;
		ant_rotb_proc = ant_rotb_proc-da;
		if (ant_rotb_proc < earth_angle_proc)ant_rotb_proc = earth_angle_proc;
	}

	SetAnimation (anim_ant_rota, ant_rota_proc);
	SetAnimation (anim_ant_rotb, ant_rotb_proc);
	}
		
	///////////// booms
	if (boom_status == 1) 
	{
		double da = SimDT*BOOM_SPEED;
		mag_boom_proc = mag_boom_proc+da;
		if (mag_boom_proc > 1)mag_boom_proc = 1;

		if (grs_boom_proc < 0.2)
		{
		grs_boom_proc = grs_boom_proc+da;
		if (grs_boom_proc > 0.2)grs_boom_proc = 0.2;
		}
		else if (grs_boom_proc > 0.2)
		{
		grs_boom_proc = grs_boom_proc-da;
		if (grs_boom_proc < 0.2)grs_boom_proc = 0.2;
		}

		if (grs_boom_proc == 0.2 && mag_boom_proc == 1)boom_status = 2;
	}
	else if (boom_status == 2) 
	{
		grs_boom_proc = 0.2;
		mag_boom_proc = 1;
	}
	else if (boom_status == 3) 
	{
		double da = SimDT*BOOM_SPEED;
		mag_boom_proc = mag_boom_proc-da;
		if (mag_boom_proc < 0)mag_boom_proc = 0;

		grs_boom_proc = grs_boom_proc-da;
		if (grs_boom_proc < 0)grs_boom_proc = 0;

		if (grs_boom_proc == 0 && mag_boom_proc == 0)boom_status = 0;
	}

	else if (boom_status == 4) 
	{
		double da = SimDT*BOOM_SPEED;
		mag_boom_proc = mag_boom_proc+da;
		if (mag_boom_proc > 1)mag_boom_proc = 1;

		grs_boom_proc = grs_boom_proc+da;
		if (grs_boom_proc > 1)grs_boom_proc = 1;

		if (grs_boom_proc == 1 && mag_boom_proc == 1)boom_status = 5;
	}
	else if (boom_status == 5) 
	{
		grs_boom_proc = 1;
		mag_boom_proc = 1;
	}
	SetAnimation (anim_grs_boom, grs_boom_proc);
	SetAnimation (anim_mag_boom, mag_boom_proc);

}
//void marsobserver::clbkPostStep (double simt, double simdt, double mjd)
//{

//}

////////// JETTISONS //////////////////////////////

void marsobserver::JettisonLid()
{
	VECTOR3 ofs = lid_off;
	VECTOR3 vofs = _V(0,0,0.5);

	char name[256];
	char classname[256];
	strcpy (classname, "MarsObserver/moc_lid");
	strcpy(name, "MOC_Cover");

	
  VESSELSTATUS vs;
  GetStatus(vs);
  VECTOR3 rvel;
  Local2Rel (ofs, vs.rpos);
  GlobalRot (vofs, rvel);
  vs.rvel = vs.rvel + rvel;
  vs.vrot.x = vs.vrot.x + (90*RAD);
  vs.status = 0;
  oapiCreateVessel (name, classname, vs);

  
  lid_status = 1;
  DelMesh(1);
}

void marsobserver::GoBang()
{
pstream_lvl = 1;
timer = timer+oapiGetSimStep();
	if (timer > 0.5) 
	{
		pstream_lvl = 0;
		DelMesh(1);
		DelMesh(0);
		bang_sw = 0;
		DelThrusterGroup(THGROUP_MAIN, true);
		DelThrusterGroup(THGROUP_ATT_FORWARD, true);
		DelThrusterGroup(THGROUP_ATT_YAWLEFT, true);
		DelThrusterGroup(THGROUP_ATT_YAWRIGHT, true);
		DelThrusterGroup(THGROUP_ATT_PITCHUP, true);
		DelThrusterGroup(THGROUP_ATT_PITCHDOWN, true);
	}

}
